(** PREAMBLE *)

(** Defining the variable environment*)
type ('a, 'b) env  = ('a * 'b) list

let update (env: ('a,'b) env) a b : ('a,'b) env = (a,b) :: env

let rec lookup (env: ('a,'b) env) a = 
  match env with
  | (a',b) :: env -> if a = a' then Some b else lookup env a
  | []            -> None 


(** Type structure for parsing*)
type operator     = Add | Sub
type exp_constant = Bcon of bool | Icon of int
type expression   = Var of string | Con of exp_constant | Oapp of operator * expression * expression 
  | Lam of string * expression | Fapp of expression * expression | Let of string * expression * expression
type value        = Ival of int | Bval of bool | Closure of string * expression * (string, value) env

let to_operator op =
  match op with
  | "Add" -> Add
  | _     -> Sub

(** Type structure for lexing*)
type  token_constant  = BCON of bool | ICON of int
type  token  = LP | RP | EQ | CON of token_constant | VAR of string | ADD | SUB | LET | IN | LAM | ARR | OAPP
let   token_array : (string * token) array = [| ("let", LET); ("in", IN); ("Oapp", OAPP);
  ("fun", LAM); ("true", CON (BCON true)); ("false", CON (BCON false));|]

let token_for_string (s:string) = 
  match Array.find_opt (fun (str, _) -> str = s) token_array with
  | Some (_, tok) -> tok
  | None          -> VAR s

let string_for_token (t:token) = 
  match Array.find_opt (fun (_, tok) -> tok = t) token_array with
  | Some (str, _) -> str
  | None          -> "unknown"


(** Helper functions for type checking*)
let digit_of_char c = Char.code c - Char.code '0'
let is_digit c      = '0' <= c && c <= '9'
let is_lc_letter c  = 'a' <= c && c <= 'z'
let is_uc_letter c  = 'A' <= c && c <= 'Z'
let is_letter c     = is_lc_letter c || is_uc_letter c
let is_whitespace c = match c with
  | ' ' | '\n' |  '\t' | '\r' -> true
  | _ -> false      


(** EVALUATION *)  
let rec eval (env: (string, value) env) (exp:expression) : value = 
  match exp with
  | Var x -> 
    ( match lookup env x with
      | Some v  -> v
      | None    -> failwith "[EVAL] Unbound variable"
    ) 
  | Con (Icon n)    -> Ival n
  | Con (Bcon b)    -> Bval b
  | Lam (x,e)       -> Closure (x, e, env)
  | Oapp (o,e1,e2)  -> eval_operators o (eval env e1) (eval env e2)
  | Fapp (e1,e2)    -> eval_funcapp (eval env e1) (eval env e2)
  | Let (x,e1,e2)   -> eval (update env x (eval env e1)) e2

and eval_funcapp (v1:value) (v2:value) : value = 
  match v1 with 
  | Closure (x,e,env) -> eval (update env x v2) e
  | _ -> failwith "[EVAL] Expected function but found none"
  
and eval_operators o v1 v2 = 
    match o, v1, v2 with
    | Add, Ival n1, Ival n2 when n1 <= max_int - n2 -> Ival (n1 + n2)
    | Sub, Ival n1, Ival n2 when n1 >= min_int + n2 -> Ival (n1 - n2)
    | _ , _ , _ -> failwith "[EVAL] Incompatible types or sizes for operation or unknown operator"

(** PARSING *)
let rec parse' (t_list:token list) : expression * token list =
  match t_list with
  | LET::VAR x::EQ::tl -> 
      let (e1,tl_e1) = parse' tl in
      let (e2,tl_e2) = parse' (verify IN tl_e1) in
      (Let(x,e1,e2), tl_e2)
  | LAM::VAR x::ARR::tl -> 
      let (e,rest) = parse' tl in
      (Lam(x,e), rest)
  | OAPP::VAR x::tl ->
      let (x,tl_x) = to_operator x, tl in
      let (e1,tl_e1) = parse' tl_x in
      let (e2,tl_e2) = parse' tl_e1 in
      (Oapp(x,e1,e2), tl_e2)
  | l -> parse_arithmetic l

and parse_arithmetic (t_list:token list) = 
  let (e,tl) = parse_funcapp t_list in parse_arithmetic' e tl
and parse_arithmetic' e1 (t_list:token list) = 
  match t_list with
  | ADD::tl -> let (e2,tl_e2) = parse_funcapp tl in parse_arithmetic' (Oapp (Add,e1,e2)) tl_e2
  | SUB::tl -> let (e2,tl_e2) = parse_funcapp tl in 
    ( try ( match eval [] e1, eval [] e2 with
      | x, y when x > y -> parse_arithmetic' (Oapp (Sub,e1,e2)) tl_e2
      | _               -> parse_arithmetic' (Oapp (Add,e1,e2)) tl_e2
    ) with Failure _    -> parse_arithmetic' (Oapp (Add,e1,e2)) tl_e2 )
  | _       -> (e1,t_list)

and parse_funcapp (t_list:token list) = 
  let (e,tl) = parse_primaries t_list in parse_funcapp' e tl
and parse_funcapp' e1 (t_list:token list) = 
  match t_list with
  | VAR x::tl         -> parse_funcapp' (Fapp(e1, Var x)) tl
  | CON (ICON x)::tl  -> parse_funcapp' (Fapp(e1, Con(Icon x))) tl
  | CON (BCON x)::tl  -> parse_funcapp' (Fapp(e1, Con(Bcon x))) tl
  | _ -> (e1, t_list)

and parse_primaries (t_list:token list) = 
  match t_list with
  | CON (ICON n)::tl -> (Con (Icon n), tl)
  | CON (BCON b)::tl -> (Con (Bcon b), tl)
  | VAR x::tl        -> (Var x, tl)
  | LP::tl  -> let (e,l) = parse' tl in (e, verify RP l)
  |  _      -> failwith "[PARSE] Failed on primaries"

(** Used where tokens are used as delimiters, e.g., 'in'.
Removes token if expected in that place*)
and verify (expected_t:token) (t_list:token list) = 
  match t_list with
  | [] -> failwith "[PARSE] no token left"
  | actual_t::tl ->
  if actual_t=expected_t then tl
  else failwith ("[PARSE] found token '" ^ (string_for_token actual_t) ^ 
  "' instead of '" ^ (string_for_token expected_t) ^ "'")

let parse (t_list:token list) : expression option =
  match parse' t_list with
  | e, [] -> Some e
  | _, _  -> None

  
(** LEXING *)
let lex (s:string) : token list = 

  let is_exhausted i = i >= String.length s in
  let is_next i c  = not (is_exhausted (i+1)) && s.[i+1] = c in

  let rec lex idx res =
    if is_exhausted idx then List.rev res
    else match s.[idx] with
      | c when is_whitespace c      -> lex (idx+1) res
      | c when is_digit c           -> lex_literal (idx+1) (digit_of_char c) res
      | c when is_letter c          -> lex_keyword (idx+1) 1 res
      | '-' when is_next (idx) '>'  -> lex (idx+2) (ARR::res)
      | '-' -> lex (idx+1) (SUB::res) 
      | '+' -> lex (idx+1) (ADD::res)
      | '=' -> lex (idx+1) (EQ::res)
      | '(' -> lex (idx+1) (LP::res)
      | ')' -> lex (idx+1) (RP::res)
      | a   -> failwith ("[LEX] Illegal char '" ^ (Char.escaped a) ^ "'")

  and lex_literal idx dgt res = 
    let e = ("[LEX] Cannot add digit to " ^ string_of_int dgt ^ ". Maximum allowed: " ^ string_of_int max_int) in
    if dgt >= max_int / 10 then failwith e
    else if is_exhausted idx then lex_finished_literal idx dgt res
    else match s.[idx] with
      | c when is_digit c -> lex_literal (idx+1) (10*dgt + digit_of_char c) res
      | _                 -> lex_finished_literal idx dgt res
  and lex_finished_literal idx dgt res = lex idx (CON (ICON dgt)::res)

  and lex_keyword idx len res = 
    if is_exhausted idx then lex_finished_keyword idx len res
    else match s.[idx] with
      | c when is_lc_letter c -> lex_keyword (idx+1) (len+1) res
      | _                     -> lex_finished_keyword idx len res
  and lex_finished_keyword idx len res =
    lex idx ( (token_for_string (String.sub s (idx-len) len)):: res )    

  in lex 0 []

  
(** MAIN *)
open Cohttp_lwt_unix
open Core

let random_state = Random.State.make_self_init ()

let get_tkprefix () : string =
  let random_num = 1 + (Random.State.int random_state 999999) in
  Printf.sprintf "TK-%06d Closed. Reason: " random_num

type kelvin = None | Some of int

let int_to_kelvin (n:int) : kelvin = if n >= 0 then Some n else None

let handle_temperature (n:int) = 
  let flag = String.rstrip (In_channel.read_all "./resources/flag.txt") in
  match int_to_kelvin n with
  | Some k  -> (get_tkprefix () ^ "\"" ^ string_of_int k ^ "K you say? That is totally fine. Just train harder\"")
  | None    -> (get_tkprefix () ^ "\"" ^ "Okay, that temperature is off the charts. We'll look into it. In the meantime, this might warm you up: " ^ flag ^ "\"")

let callback _conn req _body =
  let uri = Request.uri req
  in let handle_submit uri = 
    match Uri.get_query_param uri "form" with
    | Some "c36" -> (
      match Uri.get_query_param uri "temp" with
      | Some inp -> (
          match parse (lex ("let cel = " ^ (Uri.pct_decode inp) ^ " in cel + 273")) with
          | None -> "[MAIN] Parsing did not return any tokens. Uncatched error during parsing?"
          | Some exp -> (
              match eval [] exp with
              | Ival n    -> handle_temperature n
              | Bval b    -> "Expected temperature but received" ^ (if b then "true\n" else "false\n")
              | Closure _ -> "Expected temperature but received function value\n"
            )
        )
      | None -> "No complaint received"
      )
    | Some _  -> "Unknown form"
    | None    -> "Please specify form"
  in let handle_request uri =
    match Uri.path uri with
    | "/"                 -> In_channel.read_all "./resources/index.html"
    | "/main.css"         -> In_channel.read_all "./resources/main.css"
    | "/gym.jpg"          -> In_channel.read_all "./resources/gym.jpg"
    | "/fluxfit_logo.png" -> In_channel.read_all "./resources/fluxfit_logo.png"
    | "/submit"           -> handle_submit uri
    | _                   -> "Not Found"
  in let response = 
    try handle_request uri
    with Failure e -> e
  in Server.respond_string ~status:`OK ~body:response ()

let server =
  let port = 8080 in
  let conn_closed (_conn:Server.conn) = () in
  Server.create ~mode:(`TCP (`Port port)) (Server.make ~callback ~conn_closed ())
  
let () = Lwt_main.run server